/**
 * Gamecube controller to Nintendo 64 adapter
 * by Andrew Brown
 * Rewritten for N64 to HID by Peter Den Hartog
 */
 
 /**
 * Tetris code borrowed from wayneandlayne.com
 */

#include <TVout.h>
#include <fontALL.h>
#include "stc.h"
#include "platform.h"
#include "bitmaps.h"
#include "tones.h"
#include <avr/pgmspace.h>
#include "pins_arduino.h"
#include "crc_table.h"

#define N64_PIN 2
#define N64_PIN_DIR DDRD
#define N64_HIGH DDRD &= ~0x04
#define N64_LOW DDRD |= 0x04
#define N64_QUERY (PIND & 0x04)

#define DISABLE_INTRO

//BOARD_X and NXT_X must be divisable by 8!!!!!!!!
#define BOARD_X (24)
#define BOARD_Y (0)
#define NXT_X (0)
#define NXT_Y (8)

#define TILE_SIZE (4)

#define W (112)
#define H (90)

// 8 bytes of data that we get from the controller
struct {
    // bits: 0, 0, 0, start, y, x, b, a
    unsigned char data1;
    // bits: 1, L, R, Z, Dup, Ddown, Dright, Dleft
    unsigned char data2;
    char stick_x;
    char stick_y;
} N64_status;

TVout TV;
StcGame game;
char prev = 0;
char restarted = 0;

char N64_raw_dump[33]; // 1 received bit per byte


void tetrisSetup();
void tetrisLoop();
void intro();
void printboard();
void setCell(char x, char y,char bx, char by, char c, char f);

void N64_send(unsigned char *buffer, char length);
void N64_get();
void translate_raw_data();
void n64Setup();
void n64Loop();

void setup() 
{
  n64Setup();
  tetrisSetup();
}

void loop()
{
  n64Loop();
  tetrisLoop();
}

void tetrisLoop() 
{
  gameUpdate(&game);
}

void n64Loop()
{
    int i;
    unsigned char data, addr;

    // Command to send to the gamecube
    // The last bit is rumble, flip it to rumble
    // yes this does need to be inside the loop, the
    // array gets mutilated when it goes through N64_send
    unsigned char command[] = {0x01};
 
    // don't want interrupts getting in the way
    noInterrupts();
    // send those 3 bytes
    N64_send(command, 1);
    // read in data and dump it to N64_raw_dump
    N64_get();
    // end of time sensitive code
    interrupts();

    // translate the data in N64_raw_dump to something useful
    translate_raw_data();
}

void tetrisSetup(){
  int error;

  TV.begin(_NTSC,W,H);
  TV.select_font(font6x8);
  TV.delay(500);
  gameInit(&game);
}

void n64Setup()
{
  Serial.begin(115200);



  // Communication with gamecube controller on this pin
  // Don't remove these lines, we don't want to push +5V to the controller
  digitalWrite(N64_PIN, LOW);  
  pinMode(N64_PIN, INPUT);


  // Initialize the gamecube controller by sending it a null byte.
  // This is unnecessary for a standard controller, but is required for the
  // Wavebird.
  unsigned char initialize = 0x00;
  noInterrupts();
  N64_send(&initialize, 1);

  // Stupid routine to wait for the gamecube controller to stop
  // sending its response. We don't care what it is, but we
  // can't start asking for status if it's still responding
  int x;
  for (x=0; x<64; x++) {
      // make sure the line is idle for 64 iterations, should
      // be plenty.
      if (!N64_QUERY)
          x = 0;
  }

  // Query for the gamecube controller's status. We do this
  // to get the 0 point for the control stick.
  unsigned char command[] = {0x01};
  N64_send(command, 1);
  // read in data and dump it to N64_raw_dump
  N64_get();
  interrupts();
  translate_raw_data();  
}
//define these for the arduino platform
int platformInit(StcGame *gameInstance) {
  printboard();
  return 0;
}

void printboard() {
  TV.clear_screen();

  // label the next piece
  TV.printPGM(NXT_X,NXT_Y-8,PSTR("Next"));
  
  //print arduino tetris vertically allong the left
  TV.select_font(font8x8);
  TV.printPGM(0,NXT_Y+TETROMINO_SIZE*TILE_SIZE+8,PSTR("A\nrT\ndE\nuT\niR\nnI\noS"));
  TV.select_font(font6x8);
  TV.bitmap(W-39,H-40,TVoutMini);

  // label the Score:
  TV.printPGM(BOARD_X+BOARD_TILEMAP_WIDTH*TILE_SIZE+4,0,PSTR("lvl"));
  TV.printPGM(BOARD_X+BOARD_TILEMAP_WIDTH*TILE_SIZE+4,12,PSTR("Score"));
  TV.printPGM(BOARD_X+BOARD_TILEMAP_WIDTH*TILE_SIZE+4,28,PSTR("Lines"));

  //draw the game boarders
  TV.draw_line(BOARD_X-2,BOARD_Y, BOARD_X-2, BOARD_Y+BOARD_TILEMAP_HEIGHT*TILE_SIZE, WHITE);
  TV.draw_line(BOARD_X+BOARD_TILEMAP_WIDTH*TILE_SIZE,
               BOARD_Y,BOARD_X+BOARD_TILEMAP_WIDTH*TILE_SIZE,
               BOARD_Y+BOARD_TILEMAP_HEIGHT*TILE_SIZE,
               WHITE);
  TV.draw_line(BOARD_X-2,
               BOARD_Y+BOARD_TILEMAP_HEIGHT*TILE_SIZE+1,
               BOARD_X+BOARD_TILEMAP_WIDTH*TILE_SIZE,
               BOARD_Y+BOARD_TILEMAP_HEIGHT*TILE_SIZE+1,
               WHITE);
}
void platformEnd(StcGame *gameInstance) {
}

//modify this to read controller inputs
//there is something wrong in here but there is temp fix at the end
void platformReadInput(StcGame *gameInstance){

#define LEFTBIT 1
#define RIGHTBIT 2
#define UPBIT 4
#define DOWNBIT 8
#define ABIT 16
#define STARTBIT 32
#define SELECTBIT 64

#define PLEFT (prev &LEFTBIT)
#define PRIGHT (prev & RIGHTBIT)
#define PDOWN (prev &DOWNBIT)
#define PUP (prev & UPBIT)
#define PA (prev & ABIT)
#define PSTART (prev & STARTBIT)
#define PSELECT (prev & SELECTBIT)

  
  if ((int)N64_status.data2 == 16) {
    if (!PUP) {
      gameOnKeyDown(&game, EVENT_ROTATE_CW);
      prev |= UPBIT;
    }
  }
  else if (PUP){
    prev &= ~UPBIT;
  }
  
  if (((int)N64_status.data1) == 16) {
    if (!PSTART) {
      gameOnKeyDown(&game, EVENT_PAUSE);
      prev |= STARTBIT;
    }
  }
  else if (PSTART){
    prev &= ~STARTBIT;
  }
  
  if (((int)N64_status.data1) == 32) {
    if (!PSELECT) {
      gameOnKeyDown(&game, EVENT_RESTART);
      prev |= SELECTBIT;
    }
  }
  else if (PSELECT){
    prev &= ~SELECTBIT;
  }
  
  if (((int)N64_status.data1) == 64) {
    if (!PA) {
      gameOnKeyDown(&game, EVENT_DROP);
      prev |= ABIT;
    }
  }
  else if (PA) {
    prev &= ~ABIT;
  }
  
  if (((int)N64_status.data1) == 2) {
    if (!PLEFT) {
      gameOnKeyDown(&game, EVENT_MOVE_LEFT);
      prev |= LEFTBIT;
    }
  }
  else if (PLEFT) {
    gameOnKeyUp(&game, EVENT_MOVE_LEFT);
    prev &= ~LEFTBIT;
  }
  
  if (((int)N64_status.data1) == 1) {
    if (!PRIGHT) {
      gameOnKeyDown(&game, EVENT_MOVE_RIGHT);
      prev |= RIGHTBIT;
    }
  }
  else if (PRIGHT) {
    gameOnKeyUp(&game, EVENT_MOVE_RIGHT);
    prev &= ~RIGHTBIT;
  }
  
  if (((int)N64_status.data1) == 4) {
    if (!PDOWN) {
      gameOnKeyDown(&game, EVENT_MOVE_DOWN);
      prev |= DOWNBIT;
    }
  }
  else if (PDOWN) {
    gameOnKeyUp(&game, EVENT_MOVE_DOWN);
    prev &= ~DOWNBIT;
  }

}

  

void setCell(char x, char y,char bx, char by, char c, char f) {
  int index = bx/8 + x/2 + by*W/8 + y*4*W/8;
  if (c) {
    if (x & 1) {
      TV.screen[index] &= 0b11110000;
      TV.screen[index] |= 0b00001110;
      index+=W/8;
      TV.screen[index] &= 0b11110000;
      TV.screen[index] |= 0b00001010;
      if (f) TV.screen[index] &= 0b11111011;
      index+=W/8;
      TV.screen[index] &= 0b11110000;
      TV.screen[index] |= 0b00001110;
      index+=W/8;
      TV.screen[index] &= 0b11110000;
    }
    else {
      TV.screen[index] &= 0b00001111;
      TV.screen[index] |= 0b11100000;
      index+=W/8;
      TV.screen[index] &= 0b00001111;
      TV.screen[index] |= 0b10100000;
      if (f) TV.screen[index] &= 0b10111111;
      index+=W/8;
      TV.screen[index] &= 0b00001111;
      TV.screen[index] |= 0b11100000;
      index+=W/8;
      TV.screen[index] &= 0b00001111;
    }
  }
  else {
    if (x & 1) {
      TV.screen[index] &= 0b11110000;
      index+=W/8;
      TV.screen[index] &= 0b11110000;
      if (f) TV.screen[index] |= 0b00000100;
      index+=W/8;
      TV.screen[index] &= 0b11110000;
      index+=W/8;
      TV.screen[index] &= 0b11110000;
    }
    else {
      TV.screen[index] &= 0b00001111;
      index+=W/8;
      TV.screen[index] &= 0b00001111;
      if (f) TV.screen[index] |= 0b01000000;
      index+=W/8;
      TV.screen[index] &= 0b00001111;
      index+=W/8;
      TV.screen[index] &= 0b00001111;
    }
  }
}

void update_music()
{
  static unsigned char current_note = 0;
  static long next_note_start_time = 0;
  if (TV.millis() >= next_note_start_time)
  {
    if (current_note == SONG_LENGTH)
    {
       //it's the end of the song!
       current_note = 0;    
    }
    
      int note_duration = (pgm_read_word_near(duration_t1_c1 + current_note)/16)*5;
      next_note_start_time = TV.millis()+note_duration+PAUSE_BETWEEN_NOTES;
      TV.tone(pgm_read_word_near(melody_t1_c1 + current_note), note_duration); 
    if (current_note == SONG_LENGTH-1)
    {
      next_note_start_time += PAUSE_BETWEEN_SONGS;
      next_note_start_time -= PAUSE_BETWEEN_NOTES;
    }
    current_note++;
    
  }
}


//modify this to render the game TVout redering is defined here.
void platformRenderGame(StcGame *gameInstance) {
  char i,j;
  TV.delay_frame(1);
if (!game.isPaused && !game.isOver) {
    //draw the static blocks
    for (i = 0; i < BOARD_TILEMAP_WIDTH; i++) {
      for (j = 0; j < BOARD_TILEMAP_HEIGHT; j++) {
#ifdef STC_SHOW_GHOST_PIECE
        setCell(i,j,BOARD_X,BOARD_Y,game.map[i][j],0);
#else
        setCell(i,j,BOARD_X,BOARD_Y,game.map[i][j],1);
#endif
      }
    }
    update_music();
  
#ifdef STC_SHOW_GHOST_PIECE
    for (i = 0; i < TETROMINO_SIZE; i++) {
      for (j = 0; j < TETROMINO_SIZE; j++) {
        if (game.fallingBlock.cells[i][j] != EMPTY_CELL)
          setCell(game.fallingBlock.x + i, game.fallingBlock.y + game.shadowGap + j, BOARD_X, BOARD_Y, 0, 1);
       }
     }
#endif

    // falling block
    for (i = 0; i < TETROMINO_SIZE; i++) {
      for (j = 0; j < TETROMINO_SIZE; j++) {
        if (game.fallingBlock.cells[i][j]) {
          setCell(game.fallingBlock.x + i,game.fallingBlock.y + j,BOARD_X,BOARD_Y,1,1);
        }
      }
    }

    // nextblock block
    TV.draw_rect(NXT_X,NXT_Y,TETROMINO_SIZE*TILE_SIZE,TETROMINO_SIZE*TILE_SIZE,BLACK,BLACK);
    for (i = 0; i < TETROMINO_SIZE; i++) {
      for (j = 0; j < TETROMINO_SIZE; j++) {
        setCell (game.nextBlock.x + i+1,game.nextBlock.y + j+1,NXT_X,NXT_Y,game.nextBlock.cells[i][j],0);
      }
    }
  
    //update the stats
    TV.print(BOARD_X+BOARD_TILEMAP_WIDTH*TILE_SIZE+26,0,game.stats.level);
    TV.print(BOARD_X+BOARD_TILEMAP_WIDTH*TILE_SIZE+10,20,game.stats.score);
    TV.print(BOARD_X+BOARD_TILEMAP_WIDTH*TILE_SIZE+10,36,game.stats.lines);
  }
  
  if (game.isPaused)
    TV.printPGM(BOARD_X+4,BOARD_Y+40,PSTR("Pause"));
  if (game.isOver) {
    TV.printPGM(BOARD_X+7,BOARD_Y+36,PSTR("Game"));
    TV.printPGM(BOARD_X+7,BOARD_Y+44,PSTR("Over"));
    restarted = 1;
  }
  
  if (restarted == 1 && game.isOver == 0) {
    printboard();
    restarted = 0;
  }
  
  game.stateChanged = 0;
 
}

//return millis for this since this uses TVout we need to return TVouts version of millis.
long platformGetSystemTime() {
  return TV.millis();
}

void platformSeedRandom(long seed) {
  srand(analogRead(0)*seed);
}

int platformRandom() {
  return rand();
}

void translate_raw_data()
{
    // The get_N64_status function sloppily dumps its data 1 bit per byte
    // into the get_status_extended char array. It's our job to go through
    // that and put each piece neatly into the struct N64_status
    int i;
    memset(&N64_status, 0, sizeof(N64_status));
    // line 1
    // bits: A, B, Z, Start, Dup, Ddown, Dleft, Dright
    for (i=0; i<8; i++) {
        N64_status.data1 |= N64_raw_dump[i] ? (0x80 >> i) : 0;
    }
    // line 2
    // bits: 0, 0, L, R, Cup, Cdown, Cleft, Cright
    for (i=0; i<8; i++) {
        N64_status.data2 |= N64_raw_dump[8+i] ? (0x80 >> i) : 0;
    }
    // line 3
    // bits: joystick x value
    // These are 8 bit values centered at 0x80 (128)
    for (i=0; i<8; i++) {
        N64_status.stick_x |= N64_raw_dump[16+i] ? (0x80 >> i) : 0;
    }
    for (i=0; i<8; i++) {
        N64_status.stick_y |= N64_raw_dump[24+i] ? (0x80 >> i) : 0;
    }
}

/**
 * This sends the given byte sequence to the controller
 * length must be at least 1
 * Oh, it destroys the buffer passed in as it writes it
 */
void N64_send(unsigned char *buffer, char length)
{
    // Send these bytes
    char bits;
    
    bool bit;

    // This routine is very carefully timed by examining the assembly output.
    // Do not change any statements, it could throw the timings off
    //
    // We get 16 cycles per microsecond, which should be plenty, but we need to
    // be conservative. Most assembly ops take 1 cycle, but a few take 2
    //
    // I use manually constructed for-loops out of gotos so I have more control
    // over the outputted assembly. I can insert nops where it was impossible
    // with a for loop
    
    asm volatile (";Starting outer for loop");
outer_loop:
    {
        asm volatile (";Starting inner for loop");
        bits=8;
inner_loop:
        {
            // Starting a bit, set the line low
            asm volatile (";Setting line to low");
            N64_LOW; // 1 op, 2 cycles

            asm volatile (";branching");
            if (*buffer >> 7) {
                asm volatile (";Bit is a 1");
                // 1 bit
                // remain low for 1us, then go high for 3us
                // nop block 1
                asm volatile ("nop\nnop\nnop\nnop\nnop\n");
                
                asm volatile (";Setting line to high");
                N64_HIGH;

                // nop block 2
                // we'll wait only 2us to sync up with both conditions
                // at the bottom of the if statement
                asm volatile ("nop\nnop\nnop\nnop\nnop\n"  
                              "nop\nnop\nnop\nnop\nnop\n"  
                              "nop\nnop\nnop\nnop\nnop\n"  
                              "nop\nnop\nnop\nnop\nnop\n"  
                              "nop\nnop\nnop\nnop\nnop\n"  
                              "nop\nnop\nnop\nnop\nnop\n"  
                              );

            } else {
                asm volatile (";Bit is a 0");
                // 0 bit
                // remain low for 3us, then go high for 1us
                // nop block 3
                asm volatile ("nop\nnop\nnop\nnop\nnop\n"  
                              "nop\nnop\nnop\nnop\nnop\n"  
                              "nop\nnop\nnop\nnop\nnop\n"  
                              "nop\nnop\nnop\nnop\nnop\n"  
                              "nop\nnop\nnop\nnop\nnop\n"  
                              "nop\nnop\nnop\nnop\nnop\n"  
                              "nop\nnop\nnop\nnop\nnop\n"  
                              "nop\n");

                asm volatile (";Setting line to high");
                N64_HIGH;

                // wait for 1us
                asm volatile ("; end of conditional branch, need to wait 1us more before next bit");
                
            }
            // end of the if, the line is high and needs to remain
            // high for exactly 16 more cycles, regardless of the previous
            // branch path

            asm volatile (";finishing inner loop body");
            --bits;
            if (bits != 0) {
                // nop block 4
                // this block is why a for loop was impossible
                asm volatile ("nop\nnop\nnop\nnop\nnop\n"  
                              "nop\nnop\nnop\nnop\n");
                // rotate bits
                asm volatile (";rotating out bits");
                *buffer <<= 1;

                goto inner_loop;
            } // fall out of inner loop
        }
        asm volatile (";continuing outer loop");
        // In this case: the inner loop exits and the outer loop iterates,
        // there are /exactly/ 16 cycles taken up by the necessary operations.
        // So no nops are needed here (that was lucky!)
        --length;
        if (length != 0) {
            ++buffer;
            goto outer_loop;
        } // fall out of outer loop
    }

    // send a single stop (1) bit
    // nop block 5
    asm volatile ("nop\nnop\nnop\nnop\n");
    N64_LOW;
    // wait 1 us, 16 cycles, then raise the line 
    // 16-2=14
    // nop block 6
    asm volatile ("nop\nnop\nnop\nnop\nnop\n"
                  "nop\nnop\nnop\nnop\nnop\n"  
                  "nop\nnop\nnop\nnop\n");
    N64_HIGH;

}

void N64_get()
{
    // listen for the expected 8 bytes of data back from the controller and
    // blast it out to the N64_raw_dump array, one bit per byte for extra speed.
    // Afterwards, call translate_raw_data() to interpret the raw data and pack
    // it into the N64_status struct.
    asm volatile (";Starting to listen");
    unsigned char timeout;
    char bitcount = 32;
    char *bitbin = N64_raw_dump;

    // Again, using gotos here to make the assembly more predictable and
    // optimization easier (please don't kill me)
read_loop:
    timeout = 0x3f;
    // wait for line to go low
    while (N64_QUERY) {
        if (!--timeout)
            return;
    }
    // wait approx 2us and poll the line
    asm volatile (
                  "nop\nnop\nnop\nnop\nnop\n"  
                  "nop\nnop\nnop\nnop\nnop\n"  
                  "nop\nnop\nnop\nnop\nnop\n"  
                  "nop\nnop\nnop\nnop\nnop\n"  
                  "nop\nnop\nnop\nnop\nnop\n"  
                  "nop\nnop\nnop\nnop\nnop\n"  
            );
    *bitbin = N64_QUERY;
    ++bitbin;
    --bitcount;
    if (bitcount == 0)
        return;

    // wait for line to go high again
    // it may already be high, so this should just drop through
    timeout = 0x3f;
    while (!N64_QUERY) {
        if (!--timeout)
            return;
    }
    goto read_loop;

}
